﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using PdfSharp.Pdf.Advanced;

namespace PdfSharp.Pdf.IO
{
    /// <summary>
    /// Reads compressed objects from ObjectStreams.
    /// See Pdf reference 1.7, Section 7.5.7
    /// </summary>
    internal class ObjectStreamReader
    {
        private readonly PdfDocument document;
        private readonly Parser parser;
        private Lexer lexer;

        public ObjectStreamReader(PdfDocument doc, Parser parser, Lexer lexer)
        {
            document = doc;
            this.parser = parser;
            this.lexer = lexer;
        }

        public void UncompressObjects()
        {
            var streams = new Dictionary<int, ObjectStream>();

            document.irefTable.IsUnderConstruction = true;
            foreach (var co in document.compressedObjects)
            {
                var streamId = co.ContainingObjectId;
                if (!streams.ContainsKey(streamId))
                {
                    if (!document.irefTable.Contains(new PdfObjectID(streamId)))
                        Debug.WriteLine(String.Format("Stream {0} not found", streamId));
                    else
                    {
                        var objId = new PdfObjectID(streamId);
                        var streamObject = document.irefTable[objId];
                        if (streamObject.Value == null)
                        {
                            parser.MoveToObject(objId);
                            var obj = parser.ReadObject(null, objId, false);
                            if (obj != null)
                            {
                                try
                                {
                                    document.SecurityHandler.EncryptObject(obj);
                                }
                                catch (Exception ex)
                                {
                                    Debug.WriteLine(String.Format("Cannot decrypt ObjectStream {0}: {1}", obj.ObjectNumber, ex.Message));
                                }
                                streamObject.Value = obj;
                                var streamDict = streamObject.Value as PdfDictionary;
                                new ObjectStream(streamDict, streams);
                            }
                        }
                    }
                }
                if (streams.ContainsKey(streamId))
                {
                    // add uncompressed objects back to the xref table
                    var obj = streams[streamId].Objects[co.IndexInContainer];
                    if (obj.Reference == null)
                        document.irefTable.Add(new PdfReference(obj));
                    else
                        document.irefTable.Add(obj.Reference);
                }
                else
                    Debug.WriteLine(String.Format("Stream {0} not found", streamId));
            }
            // Remove Object streams
            foreach (var kv in streams)
                document.irefTable.objectTable.Remove(new PdfObjectID(kv.Key));

            document.compressedObjects.Clear();
            document.irefTable.IsUnderConstruction = false;
            FixReferences();
        }

        private void FixReferences()
        {
            var refs = document.irefTable.AllReferences;
            foreach (var r in refs)
            {
                if (r.Value is PdfDictionary)
                    FixDictionary(r.Value as PdfDictionary);
                else if (r.Value is PdfArray)
                    FixArray(r.Value as PdfArray);
            }
        }

        private void FixDictionary(PdfDictionary dict)
        {
            var keys = dict.Elements.KeyNames;
            foreach (var key in keys)
            {
                var obj = dict.Elements[key];
                if (obj is PdfDictionary)
                    FixDictionary(obj as PdfDictionary);
                else if (obj is PdfArray)
                    FixArray(obj as PdfArray);
                else if (obj is PdfReference)
                {
                    var r = (PdfReference)obj;
                    if (r.Value == null)
                    {
                        var iref = document.irefTable[r.ObjectID];
                        dict.Elements[key] = iref;
                    }
                }
            }
        }

        private void FixArray(PdfArray array)
        {
            for (var i = 0; i < array.Elements.Count; i++)
            {
                var obj = array.Elements[i];
                if (obj is PdfDictionary)
                    FixDictionary(obj as PdfDictionary);
                else if (obj is PdfArray)
                    FixArray(obj as PdfArray);
                else if (obj is PdfReference)
                {
                    var r = (PdfReference)obj;
                    if (r.Value == null)
                    {
                        var iref = document.irefTable[r.ObjectID];
                        array.Elements[i] = iref;
                    }
                }
            }
        }

        private class ObjectStream
        {
            /// <summary>
            /// Gets the list of objects contained in this ObjectStream
            /// </summary>
            public List<PdfObject> Objects { get; private set; }

            /// <summary>
            /// Gets the ObjectStream this ObjectStream is an extension of
            /// </summary>
            public ObjectStream BaseStream { get; private set; }

            private readonly PdfDictionary stream;

            private Parser parser;

            public ObjectStream(PdfDictionary stream, Dictionary<int, ObjectStream> streams)
            {
                this.stream = stream;
                Objects = new List<PdfObject>();
                ReadRecursively(streams);
            }

            private void ReadRecursively(Dictionary<int, ObjectStream> streams)
            {
                var data = stream.Stream.UnfilteredValue;
                var ms = new MemoryStream(data);
                parser = new Parser(stream.document, ms);
                Read();
                streams.Add(stream.ObjectNumber, this);
                // i have currently no idea what to do with the base stream...
                var extends = stream.Elements.GetReference("/Extends");
                if (extends != null)
                {
                    if (!streams.ContainsKey(extends.ObjectNumber))
                    {
                        Debug.WriteLine(String.Format("Stream {0} extends Stream {1}", stream.ObjectNumber, extends.ObjectNumber));
                        var o = stream.document.irefTable[extends.ObjectID];
                        if (o != null)
                        {
                            BaseStream = new ObjectStream(o.Value as PdfDictionary, streams);
                        }
                        else
                            Debug.WriteLine(String.Format("Stream {0} extends undefined Stream {1}", stream.ObjectNumber, extends.ObjectNumber));
                    }
                    else
                        BaseStream = streams[extends.ObjectNumber];
                }
            }

            private void Read()
            {
                var lexer = parser.Lexer;
                var objectCount = stream.Elements.GetInteger("/N");
                var startOffset = stream.Elements.GetInteger("/First");
                var objectIds = new int[objectCount];
                var objectOffsets = new int[objectCount];
                for (var i = 0; i < objectCount; i++)
                {
                    lexer.ScanNextToken();
                    objectIds[i] = lexer.TokenToInteger;
                    lexer.ScanNextToken();
                    objectOffsets[i] = lexer.TokenToInteger + startOffset;
                }
                for (var i = 0; i < objectCount; i++)
                {
                    lexer.Position = objectOffsets[i];
                    var pdfObject = parser.ReadObject(null, new PdfObjectID(objectIds[i]), false, true);
                    pdfObject.FromObjectStream = true;
                    Objects.Add(pdfObject);
                }
            }
        }
    }
}