﻿using PdfSharp.Pdf.Advanced;
using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace PdfSharp.Pdf.IO
{
    class XrefStreamReader
    {
        private readonly PdfDocument document;
        private readonly Parser parser;
        private readonly Lexer lexer;
        private int recursions;

        public XrefStreamReader(PdfDocument document, Parser parser, Lexer lexer)
        {
            this.document = document;
            this.parser = parser;
            this.lexer = lexer;
            recursions = 0;
        }

        public PdfTrailer Read()
        {
            var trailer = new PdfTrailer(document);
            var objId = lexer.TokenToInteger;
            lexer.ScanNextToken();
            Debug.Assert(lexer.Symbol == Symbol.Integer, "Expected an Integer");
            var generation = lexer.TokenToInteger;
            lexer.ScanNextToken();  // obj
            lexer.ScanNextToken();  // begin dictionary
            var dict = parser.ReadDictionary(trailer, false);
            Debug.Assert(dict.Elements.GetName("/Type") == "/XRef", "Expected a dictionary of type XRef");
            lexer.ScanNextToken();  // begin stream
            Debug.Assert(lexer.Symbol == Symbol.BeginStream, "Expected a Stream");
            var length = parser.GetStreamLength(dict);
            var bytes = lexer.ReadStream(length);
            lexer.ScanNextToken();  // endstream
            dict.CreateStream(bytes);
            bytes = dict.Stream.UnfilteredValue;
            var decodeParams = dict.Elements.GetDictionary("/DecodeParms");
            if (decodeParams != null)
            {
                var width = decodeParams.Elements.GetInteger("/Columns");
                var predictor = decodeParams.Elements.GetInteger("/Predictor");
                if (predictor != 12)
                    throw new PdfReaderException("Unsupported Predictor " + predictor + ". Only Type 12 is supported");
                if (bytes.Length % (width + 1) != 0)
                    throw new PdfSharpException(String.Format("Length of Xref Stream ({0}) is not a multiple of {1}", bytes.Length, width + 1));
                bytes = Unpredict(bytes, width);
            }
            var lastObjectIdPlusOne = dict.Elements.GetInteger("/Size");

            // See Pdf Reference 1.7, Section 7.5.8.2
            var sections = new [] { 0, lastObjectIdPlusOne };    // assume single section
            var index = dict.Elements.GetArray("/Index");
            if (index != null)
            {
                sections = new int[index.Elements.Count];
                for (var i = 0; i < sections.Length; i++)
                    sections[i] = index.Elements.GetInteger(i);
            }
            var columns = dict.Elements.GetArray("/W");
            if (columns.Elements.Count != 3)
                throw new PdfReaderException("Expected W-Array to be of length 3. Instead it was of length " + columns.Elements.Count);
            var colWidths = new int[columns.Elements.Count];
            for (var i = 0; i < colWidths.Length; i++)
                colWidths[i] = columns.Elements.GetInteger(i);

            // See Pdf Reference 1.7, Section 7.5.8.3
            var sectionIndex = 0;
            var streamIndex = 0;
            do
            {
                var curObjectId = sections[sectionIndex];
                var objCount = sections[sectionIndex + 1];
                var numRead = 0;
                while (numRead < objCount)
                {
                    var type = ReadInteger(colWidths[0], bytes, streamIndex);
                    streamIndex += colWidths[0];
                    var fileOffset = ReadInteger(colWidths[1], bytes, streamIndex);
                    streamIndex += colWidths[1];
                    var objGen = ReadInteger(colWidths[2], bytes, streamIndex);
                    streamIndex += colWidths[2];
                    numRead++;
                    var curObjId = curObjectId > 0 ? new PdfObjectID(curObjectId) : PdfObjectID.Empty;
                    if (type == 1)      // normal object
                    {
                        if (!document.irefTable.Contains(curObjId))
                            document.irefTable.Add(new PdfReference(curObjId, fileOffset));
                    }
                    else if (type == 2) // compressed object
                    {
                        var co = new PdfCompressedObject(curObjectId, fileOffset, objGen);
                        document.compressedObjects.Add(co);
                    }
                    curObjectId++;
                }
                sectionIndex += 2;
            } while (sectionIndex < sections.Length);

            if (dict.Elements.ContainsKey(PdfTrailer.Keys.Prev))
            {
                var prev = dict.Elements.GetInteger(PdfTrailer.Keys.Prev);
                lexer.Position = prev;
                lexer.ScanNextToken();
                recursions++;
                Read();
                recursions--;
            }

            if (recursions == 0)
            {
                trailer.Elements.Remove(PdfTrailer.Keys.Prev);
                trailer.Elements.Remove("/DecodeParms");
                trailer.Elements.Remove("/Filter");
                trailer.Elements.Remove("/Index");
                trailer.Elements.Remove("/W");
            }

            return trailer;
        }

        private byte[] Unpredict(byte[] data, int rowWidth)
        {
            var bytes = new List<byte>();
            var prevRow = new byte[rowWidth];
            var row = new byte[rowWidth + 1];
            var pos = 0;
            while (pos < data.Length)
            {
                Array.Copy(data, pos, row, 0, rowWidth + 1);
                if (row[0] != 2)
                    throw new PdfReaderException(String.Format("Unexpected Prefix {0} in Xref Stream Row. Expected '2'.", row[0]));
                for (var i = 1; i < row.Length; i++)
                {
                    var b = row[i];
                    b += prevRow[i - 1];
                    prevRow[i - 1] = b;
                    bytes.Add(b);
                }
                pos += rowWidth + 1;
            }
            return bytes.ToArray();
        }

        private static int ReadInteger(int count, byte[] data, int offset)
        {
            var result = 0;
            for (var i = 0; i < count; i++)
            {
                result <<= 8;
                result += data[offset + i];
            }
            return result;
        }
    }
}
