﻿using System;
using System.IO;
using System.Security.Cryptography;
using System.Text;

namespace PdfSharp.Pdf.Security
{
    internal class AESEncryptor : RC4Encryptor, IEncryptor
    {
        public override void InitEncryptionKey(string password)
        {
            if (rValue == 5)
            {
                InitVersion5(password);
                return;
            }
            if (rValue == 6)
            {
                // http://esec-lab.sogeti.com/post/The-undocumented-password-validation-algorithm-of-Adobe-Reader-X
                // Reference Implementation: http://www.mupdf.com/docs/browse/source/pdf/pdf-crypt.c#459
                InitVersion6(password);
                return;
            }

            var userPad = PadPassword(password);
            encryptionKey = new byte[vValue >= 3 ? 16 : 5];

            md5.Initialize();
            md5.TransformBlock(userPad, 0, userPad.Length, userPad, 0);
            md5.TransformBlock(ownerValue, 0, ownerValue.Length, ownerValue, 0);

            // Split permission into 4 bytes
            var permission = new byte[4];
            permission[0] = (byte)pValue;
            permission[1] = (byte)(pValue >> 8);
            permission[2] = (byte)(pValue >> 16);
            permission[3] = (byte)(pValue >> 24);
            md5.TransformBlock(permission, 0, 4, permission, 0);
            md5.TransformBlock(documentId, 0, documentId.Length, documentId, 0);
            if (rValue >= 4 && !encryptMetadata)
            {
                var ff = new byte[] { 0xff, 0xff, 0xff, 0xff };
                md5.TransformBlock(ff, 0, ff.Length, ff, 0);
            }
            md5.TransformFinalBlock(permission, 0, 0);
            var digest = md5.Hash;
            md5.Initialize();
            // Create the hash 50 times (only for 128 bit)
            if (rValue >= 3)
            {
                for (var idx = 0; idx < 50; idx++)
                {
                    digest = md5.ComputeHash(digest);
                    md5.Initialize();
                }
            }
            Array.Copy(digest, 0, encryptionKey, 0, encryptionKey.Length);

            CreateOwnerKey(password);
            CreateUserKey(password);
        }

        /// <summary>
        /// Pdf Reference 1.7 Extension Level 3, Chapter 3.5.2, Algorithm 3.2a
        /// </summary>
        /// <param name="password"></param>
        protected void InitVersion5(string password)
        {
            // TODO: apply SASLprep to the password (RFC 4013, RFC 3454)
            if (password.Length > 127)
                password = password.Substring(0, 127);
            // split O and U into their components
            var oHash = new byte[32];
            var oValidation = new byte[8];
            var oSalt = new byte[8];
            var uHash = new byte[32];
            var uValidation = new byte[8];
            var uSalt = new byte[8];
            var aes256Cbc = new AesCryptoServiceProvider { KeySize = 256, Mode = CipherMode.CBC, Padding = PaddingMode.None };
            //var aesEcb = new AesCryptoServiceProvider { KeySize = 256, Mode = CipherMode.ECB };   // for /Perms-entry

            Array.Copy(ownerValue, oHash, 32);
            Array.Copy(ownerValue, 32, oValidation, 0, 8);
            Array.Copy(ownerValue, 40, oSalt, 0, 8);
            Array.Copy(userValue, uHash, 32);
            Array.Copy(userValue, 32, uValidation, 0, 8);
            Array.Copy(userValue, 40, uSalt, 0, 8);

            var pwdBytes = Encoding.UTF8.GetBytes(password);
            var oKeyBytes = new byte[pwdBytes.Length + 8 + 48];
            Array.Copy(pwdBytes, oKeyBytes, pwdBytes.Length);
            Array.Copy(oValidation, 0, oKeyBytes, pwdBytes.Length, 8);
            Array.Copy(userValue, 0, oKeyBytes, pwdBytes.Length + 8, 48);

            computedOwnerValue = new byte[32];
            Array.Copy(oKeyBytes, computedOwnerValue, 32);

            var haveOwnerPassword = PasswordMatchR5(oKeyBytes, ownerValue);
            if (haveOwnerPassword)
            {
                CreateEncryptionKeyR5(pwdBytes, oSalt, userValue);
                //oKeyBytes = new byte[pwdBytes.Length + 8 + 48];
                //Array.Copy(pwdBytes, oKeyBytes, pwdBytes.Length);
                //Array.Copy(oSalt, 0, oKeyBytes, pwdBytes.Length, 8);
                //Array.Copy(userValue, 0, oKeyBytes, pwdBytes.Length + 8, 48);
                //var sha = SHA256.Create();
                //var tempOwnerKey = sha.ComputeHash(oKeyBytes);
                //// TODO: Use AES256 !
                //using (var decryptor = aes256Cbc.CreateDecryptor(tempOwnerKey, new byte[] { 0 }))
                //{
                //    using (var ms = new MemoryStream(oeValue))
                //    {
                //        using (var cs = new CryptoStream(ms, decryptor, CryptoStreamMode.Read))
                //        {
                //            encryptionKey = new byte[32];
                //            cs.Read(encryptionKey, 0, 32);
                //        }
                //    }
                //}
            }
            else
            {
                oKeyBytes = new byte[pwdBytes.Length + 8];
                Array.Copy(pwdBytes, oKeyBytes, pwdBytes.Length);
                Array.Copy(uValidation, 0, oKeyBytes, pwdBytes.Length, 8);

                computedUserValue = new byte[32];
                Array.Copy(oKeyBytes, computedUserValue, 32);

                // if the result matches the first 32 bytes of ownerValue, we have the user password
                var haveUserPassword = PasswordMatchR5(oKeyBytes, userValue);
                if (haveUserPassword)
                {
                    CreateEncryptionKeyR5(pwdBytes, uSalt, null);
                    //oKeyBytes = new byte[pwdBytes.Length + 8];
                    //Array.Copy(pwdBytes, oKeyBytes, pwdBytes.Length);
                    //Array.Copy(uSalt, 0, oKeyBytes, pwdBytes.Length, 8);
                    //var sha = SHA256.Create();
                    //var tempOwnerKey = sha.ComputeHash(oKeyBytes);
                    //// TODO: Use AES256 !
                    //using (var decryptor = aes256Cbc.CreateDecryptor(tempOwnerKey, new byte[] { 0 }))
                    //{
                    //    using (var ms = new MemoryStream(ueValue))
                    //    {
                    //        using (var cs = new CryptoStream(ms, decryptor, CryptoStreamMode.Read))
                    //        {
                    //            encryptionKey = new byte[32];
                    //            cs.Read(encryptionKey, 0, 32);
                    //        }
                    //    }
                    //}
                }
            }
        }

        private void CreateEncryptionKeyR5(byte[] password, byte[] salt, byte[] uservalue)
        {
            var sha = SHA256.Create();
            var aes256Cbc = new AesCryptoServiceProvider { KeySize = 256, Mode = CipherMode.CBC, Padding = PaddingMode.None };
            var bufLen = password.Length + salt.Length + (uservalue != null ? 48 : 0);
            var buf = new byte[bufLen];
            Array.Copy(password, buf, password.Length);
            Array.Copy(salt, 0, buf, password.Length, salt.Length);
            if (uservalue != null)
                Array.Copy(uservalue, 0, buf, password.Length + salt.Length, 48);
            var key = sha.ComputeHash(buf);
            using (var decryptor = aes256Cbc.CreateDecryptor(key, new byte[] { 0 }))
            {
                using (var ms = new MemoryStream(ueValue))
                {
                    using (var cs = new CryptoStream(ms, decryptor, CryptoStreamMode.Read))
                    {
                        encryptionKey = new byte[32];
                        cs.Read(encryptionKey, 0, 32);
                    }
                }
            }
            aes256Cbc.Clear();
        }

        private void InitVersion6(string password)
        {
            // split O and U into their components
            var oSalt = new byte[8];
            var uSalt = new byte[8];
            var uKeySalt = new byte[8];
            var oKeySalt = new byte[8];
            var oKey = new byte[48];
            Array.Copy(ownerValue, 32, oSalt, 0, 8);
            Array.Copy(userValue, 32, uSalt, 0, 8);
            Array.Copy(userValue, 40, uKeySalt, 0, 8);
            Array.Copy(ownerValue, 40, oKeySalt, 0, 8);
            Array.Copy(userValue, oKey, 48);

            computedUserValue = new byte[32];
            computedOwnerValue = new byte[32];
            ValidateVersion6(password, uSalt, null, computedUserValue);
            ValidateVersion6(password, oSalt, oKey, computedOwnerValue);

            byte[] keyToDecrypt = null;
            byte[] salt = null;
            byte[] hashKey = null;
            if (CompareArrays(computedOwnerValue, ownerValue, 32))
            {
                keyToDecrypt = oeValue;
                salt = oKeySalt;
                hashKey = oKey;
            }
            else if (CompareArrays(computedUserValue, userValue, 32))
            {
                keyToDecrypt = ueValue;
                salt = uKeySalt;
            }

            if (keyToDecrypt != null)
            {
                encryptionKey = new byte[32];
                var hash = new byte[32];
                var iv = new byte[16];
                ValidateVersion6(password, salt, hashKey, hash);
                var aes256 = new AesCryptoServiceProvider { KeySize = 256, Mode = CipherMode.CBC, Padding = PaddingMode.None };
                using (var decryptor = aes256.CreateDecryptor(hash, iv))
                {
                    decryptor.TransformBlock(keyToDecrypt, 0, 32, encryptionKey, 0);
                }
                aes256.Clear();
            }
        }

        // straight conversion from http://www.mupdf.com/docs/browse/source/pdf/pdf-crypt.c#459
        private void ValidateVersion6(string password, byte[] salt, byte[] ownerKey, byte[] hash)
        {
            byte[] data = new byte[(128 + 64 + 48) * 64];
            byte[] block = new byte[64];
            int blockSize = 32;
            int dataLen = 0;
            int i, j, sum;

            var sha256 = SHA256.Create();
            var sha384 = SHA384.Create();
            var sha512 = SHA512.Create();
            var aes128 = new AesCryptoServiceProvider { BlockSize = 16 * 8, Mode = CipherMode.CBC };
            var pwdBytes = Encoding.UTF8.GetBytes(password);
            var iv = new byte[16];
            var aesKey = new byte[16];

            /* Step 1: calculate initial data block */
            sha256.TransformBlock(pwdBytes, 0, pwdBytes.Length, pwdBytes, 0);
            sha256.TransformBlock(salt, 0, salt.Length, salt, 0);
            if (ownerKey != null)
                sha256.TransformBlock(ownerKey, 0, ownerKey.Length, ownerKey, 0);
            sha256.TransformFinalBlock(salt, 0, 0);
            Array.Copy(sha256.Hash, block, sha256.HashSize / 8);
            for (i = 0; i < 64 || i < data[dataLen * 64 - 1] + 32; i++)
            {
                /* Step 2: repeat password and data block 64 times */
                Array.Copy(pwdBytes, data, pwdBytes.Length);
                Array.Copy(block, 0, data, pwdBytes.Length, blockSize);
                if (ownerKey != null)
                    Array.Copy(ownerKey, 0, data, pwdBytes.Length + blockSize, 48);
                dataLen = pwdBytes.Length + blockSize + (ownerKey != null ? 48 : 0);
                for (j = 1; j < 64; j++)
                    Array.Copy(data, 0, data, j * dataLen, dataLen);

                /* Step 3: encrypt data using data block as key and iv */
                Array.Copy(block, 16, iv, 0, 16);
                Array.Copy(block, 0, aesKey, 0, 16);
                using (var aesEnc = aes128.CreateEncryptor(aesKey, iv))
                {
                    var cnt = aesEnc.TransformBlock(data, 0, dataLen * 64, data, 0);

                    /* Step 4: determine SHA-2 hash size for this round */
                    for (j = 0, sum = 0; j < 16; j++)
                        sum += data[j];

                    /* Step 5: calculate data block for next round */
                    blockSize = 32 + (sum % 3) * 16;
                    switch (blockSize)
                    {
                        case 32:
                            sha256 = SHA256.Create();
                            sha256.TransformBlock(data, 0, dataLen * 64, data, 0);
                            sha256.TransformFinalBlock(data, 0, 0);
                            Array.Copy(sha256.Hash, block, sha256.HashSize / 8);
                            break;
                        case 48:
                            sha384 = SHA384.Create();
                            sha384.TransformBlock(data, 0, dataLen * 64, data, 0);
                            sha384.TransformFinalBlock(data, 0, 0);
                            Array.Copy(sha384.Hash, block, sha384.HashSize / 8);
                            break;
                        case 64:
                            sha512 = SHA512.Create();
                            sha512.TransformBlock(data, 0, dataLen * 64, data, 0);
                            sha512.TransformFinalBlock(data, 0, 0);
                            Array.Copy(sha512.Hash, block, sha512.HashSize / 8);
                            break;
                    }
                }
            }
            Array.Copy(block, hash, 32);
        }

        private static bool PasswordMatchR5(byte[] key, byte[] comparand)
        {
            var sha = SHA256.Create();
            var hash = sha.ComputeHash(key);
            for (var i = 0; i < 32; i++)
            {
                if (hash[i] != comparand[i])
                {
                    return false;
                }
            }
            return true;
        }

        /// <summary>
        /// Pdf Reference 1.7, Chapter 7.6.2, Algorithm #1
        /// </summary>
        /// <param name="id"></param>
        public override void CreateHashKey(PdfObjectID id)
        {
            if (rValue >= 5)
            {
                key = new byte[encryptionKey.Length];
                Array.Copy(encryptionKey, key, encryptionKey.Length);
                return;
            }
            var objectId = new byte[5];
            md5.Initialize();
            // Split the object number and generation
            objectId[0] = (byte)id.ObjectNumber;
            objectId[1] = (byte)(id.ObjectNumber >> 8);
            objectId[2] = (byte)(id.ObjectNumber >> 16);
            objectId[3] = (byte)id.GenerationNumber;
            objectId[4] = (byte)(id.GenerationNumber >> 8);
            var salt = new byte[] { 0x73, 0x41, 0x6C, 0x54 };
            var k = new byte[encryptionKey.Length + 9];
            Array.Copy(encryptionKey, k, encryptionKey.Length);
            Array.Copy(objectId, 0, k, encryptionKey.Length, objectId.Length);
            Array.Copy(salt, 0, k, encryptionKey.Length + objectId.Length, salt.Length);
            key = md5.ComputeHash(k);
            md5.Initialize();
            keySize = encryptionKey.Length + 5;
            if (keySize > 16)
                keySize = 16;
        }

        public override byte[] Encrypt(byte[] bytes)
        {
            if (bytes.Length <= 16)
                return bytes;

            var iv = new byte[16];
            Array.Copy(bytes, iv, 16);
            var aes = new AesCryptoServiceProvider { Mode = CipherMode.CBC, Padding = PaddingMode.None };
            var decryptor = aes.CreateDecryptor(key, iv);
            var output = new byte[bytes.Length - 16];
            try
            {
                var offset = decryptor.TransformBlock(bytes, 16, bytes.Length - 16, output, 0);
                var suffix = decryptor.TransformFinalBlock(bytes, 0, 0);
                Array.Copy(suffix, 0, output, offset, suffix.Length);
            }
            //catch
            //{
            //    return bytes;
            //}
            finally
            {
                decryptor.Dispose();
                aes.Clear();
            }
            return output;
        }
    }
}